@@ -21,20 +21,22 @@
 #include <atomic>
 #include <cassert>
 #include <cerrno>
+#include <chrono>
 #include <cstdlib>
 #include <cstring>
 #include <fstream>
+#include <functional>
 #include <iostream>
 #include <memory>
 #include <mutex>
 #include <sstream>
 #include <thread>
-#include <chrono>
+#include <atomic>
 
-#include "Common.hpp"
-#include "Log.hpp"
-#include "Util.hpp"
-#include "SigUtil.hpp"
+#include "common/Common.hpp"
+#include "common/Log.hpp"
+#include "common/Util.hpp"
+#include "common/SigUtil.hpp"
 
 namespace Poco
 {
@@ -72,7 +74,7 @@ public:
     }
     void setMove(MoveFunction moveFn)
     {
-        _socketMove = moveFn;
+        _socketMove = std::move(moveFn);
         _disposition = Type::MOVE;
     }
     void setClosed()
@@ -133,11 +135,11 @@ public:
                             int events) = 0;
 
     /// manage latency issues around packet aggregation
-    virtual void setNoDelay()
+    void setNoDelay()
     {
         const int val = 1;
-        setsockopt (_fd, IPPROTO_TCP, TCP_NODELAY,
-                    (char *) &val, sizeof(val));
+        ::setsockopt(_fd, IPPROTO_TCP, TCP_NODELAY,
+                     (char *) &val, sizeof(val));
     }
 
     /// Sets the kernel socket send buffer in size bytes.
@@ -657,6 +659,7 @@ class StreamSocket;
 class SocketHandlerInterface
 {
 public:
+    virtual ~SocketHandlerInterface() {}
     /// Called when the socket is newly created to
     /// set the socket associated with this ResponseClient.
     /// Will be called exactly once.
@@ -694,7 +697,9 @@ public:
         Socket(fd),
         _socketHandler(std::move(socketHandler)),
         _closed(false),
-        _shutdownSignalled(false)
+        _shutdownSignalled(false),
+        _bytesSent(0),
+        _bytesRecvd(0)
     {
         LOG_DBG("StreamSocket ctor #" << fd);
 
@@ -788,6 +793,7 @@ public:
             if (len > 0)
             {
                 assert (len <= ssize_t(sizeof(buf)));
+                _bytesRecvd += len;
                 _inBuffer.insert(_inBuffer.end(), &buf[0], &buf[len]);
             }
             // else poll will handle errors.
@@ -817,6 +823,12 @@ public:
         return socket;
     }
 
+    void getIOStats(uint64_t &sent, uint64_t &recv)
+    {
+        sent = _bytesSent;
+        recv = _bytesRecvd;
+    }
+
 protected:
 
     /// Called when a polling event is received.
@@ -918,6 +930,7 @@ protected:
 
             if (len > 0)
             {
+                _bytesSent += len;
                 _outBuffer.erase(_outBuffer.begin(), _outBuffer.begin() + len);
             }
             else
@@ -958,6 +971,9 @@ protected:
     std::vector< char > _inBuffer;
     std::vector< char > _outBuffer;
 
+    uint64_t _bytesSent;
+    uint64_t _bytesRecvd;
+
     // To be able to access _inBuffer and _outBuffer.
     // TODO we probably need accessors to the _inBuffer & _outBuffer
     // instead of this many friends...
@@ -967,6 +983,21 @@ protected:
     friend class SimpleResponseClient;
 };
 
+enum class WSOpCode : unsigned char {
+    Continuation = 0x0,
+    Text         = 0x1,
+    Binary       = 0x2,
+    Reserved1    = 0x3,
+    Reserved2    = 0x4,
+    Reserved3    = 0x5,
+    Reserved4    = 0x6,
+    Reserved5    = 0x7,
+    Close        = 0x8,
+    Ping         = 0x9,
+    Pong         = 0xa
+    // ... reserved
+};
+
 namespace HttpHelper
 {
     /// Sends file as HTTP response.
