@@ -9,26 +9,29 @@
 
 #include "config.h"
 
+#include <iomanip>
 #include <string>
 #include <vector>
-#include <unistd.h>
-#include <sys/stat.h>
+
 #include <dirent.h>
+#include <sys/stat.h>
+#include <unistd.h>
 #include <zlib.h>
 
+#include <openssl/evp.h>
+
 #include <Poco/DateTime.h>
 #include <Poco/DateTimeFormat.h>
 #include <Poco/DateTimeFormatter.h>
 #include <Poco/Exception.h>
 #include <Poco/FileStream.h>
-#include <Poco/Net/HTTPCookie.h>
-#include <Poco/Net/HTTPBasicCredentials.h>
-#include <Poco/Net/HTTPResponse.h>
 #include <Poco/Net/HTMLForm.h>
+#include <Poco/Net/HTTPBasicCredentials.h>
+#include <Poco/Net/HTTPCookie.h>
 #include <Poco/Net/HTTPRequest.h>
+#include <Poco/Net/HTTPResponse.h>
 #include <Poco/Net/NameValueCollection.h>
 #include <Poco/Net/NetException.h>
-#include <Poco/RegularExpression.h>
 #include <Poco/Runnable.h>
 #include <Poco/StreamCopier.h>
 #include <Poco/StringTokenizer.h>
@@ -39,14 +42,13 @@
 #include "FileServer.hpp"
 #include "LOOLWSD.hpp"
 #include "Log.hpp"
+#include "Protocol.hpp"
 
-using Poco::FileInputStream;
 using Poco::Net::HTMLForm;
+using Poco::Net::HTTPBasicCredentials;
 using Poco::Net::HTTPRequest;
 using Poco::Net::HTTPResponse;
 using Poco::Net::NameValueCollection;
-using Poco::Net::HTTPBasicCredentials;
-using Poco::StreamCopier;
 using Poco::Util::Application;
 
 std::map<std::string, std::pair<std::string, std::string>> FileServerRequestHandler::FileHash;
@@ -77,20 +79,60 @@ bool FileServerRequestHandler::isAdminLoggedIn(const HTTPRequest& request,
         LOG_INF("No existing JWT cookie found");
     }
 
+    HTTPBasicCredentials credentials(request);
+    std::string userProvidedPwd = credentials.getPassword();
+
     // If no cookie found, or is invalid, let admin re-login
-    const auto user = config.getString("admin_console.username", "");
-    const auto pass = config.getString("admin_console.password", "");
+    const std::string user = config.getString("admin_console.username", "");
+    std::string pass = config.getString("admin_console.password", "");
+
+    if (config.has("admin_console.secure_password"))
+    {
+#if HAVE_PKCS5_PBKDF2_HMAC
+        pass = config.getString("admin_console.secure_password");
+        // Extract the salt from the config
+        std::vector<unsigned char> saltData;
+        std::vector<std::string> tokens = LOOLProtocol::tokenize(pass, '.');
+        if (tokens.size() != 5 ||
+            tokens[0] != "pbkdf2" ||
+            tokens[1] != "sha512" ||
+            !Util::dataFromHexString(tokens[3], saltData))
+        {
+            LOG_ERR("Incorrect format detected for secure_password in config file."
+                    << "Denying access until correctly set."
+                    << "Use loolconfig to configure admin password.");
+            return false;
+        }
+
+        unsigned char userProvidedPwdHash[tokens[4].size() / 2];
+        PKCS5_PBKDF2_HMAC(userProvidedPwd.c_str(), -1,
+                          saltData.data(), saltData.size(),
+                          std::stoi(tokens[2]),
+                          EVP_sha512(),
+                          sizeof userProvidedPwdHash, userProvidedPwdHash);
+
+        std::stringstream stream;
+        for (unsigned long j = 0; j < sizeof userProvidedPwdHash; ++j)
+            stream << std::hex << std::setw(2) << std::setfill('0') << static_cast<int>(userProvidedPwdHash[j]);
+
+        userProvidedPwd = stream.str();
+        pass = tokens[4];
+#else
+        LOG_ERR("The config file has admin_console.secure_password setting, "
+                << "but this application was compiled with old OpenSSL version, "
+                << "and this setting cannot be used. Falling back to plain text password, if it is set.");
+#endif
+    }
+
     if (user.empty() || pass.empty())
     {
         LOG_ERR("Admin Console credentials missing. Denying access until set.");
         return false;
     }
 
-    HTTPBasicCredentials credentials(request);
     if (credentials.getUsername() == user &&
-        credentials.getPassword() == pass)
+        userProvidedPwd == pass)
     {
-        const std::string htmlMimeType = "text/html";
         // generate and set the cookie
         JWTAuth authAgent(sslKeyPath, "admin", "admin", "admin");
         const std::string jwtToken = authAgent.getAccessToken();
@@ -122,10 +164,10 @@ void FileServerRequestHandler::handleRequest(const HTTPRequest& request, Poco::M
 
         std::vector<std::string> requestSegments;
         requestUri.getPathSegments(requestSegments);
-        if (requestSegments.size() < 1)
-        {
+        const std::string relPath = getRequestPathname(request);
+        // Is this a file we read at startup - if not; its not for serving.
+        if (requestSegments.size() < 1 || FileHash.find(relPath) == FileHash.end())
             throw Poco::FileNotFoundException("Invalid URI request: [" + requestUri.toString() + "].");
-        }
 
         const auto& config = Application::instance().config();
         const std::string loleafletHtml = config.getString("loleaflet_html", "loleaflet.html");
@@ -153,11 +195,6 @@ void FileServerRequestHandler::handleRequest(const HTTPRequest& request, Poco::M
                 response.add("Referrer-Policy", "no-referrer");
             }
 
-            const std::string relPath = getRequestPathname(request);
-            // Is this a file we read at startup - if not; its not for serving.
-            if (FileHash.find(relPath) == FileHash.end())
-                throw Poco::FileAccessDeniedException("Invalid or forbidden file path: [" + relPath + "].");
-
             // Do we have an extension.
             const std::size_t extPoint = endPoint.find_last_of('.');
             if (extPoint == std::string::npos)
@@ -291,7 +328,7 @@ void FileServerRequestHandler::readDirToHash(const std::string &basePath, const
     struct stat fileStat;
     DIR *workingdir;
 
-    LOG_TRC("Pre-reading directory: " << basePath + path << "\n");
+    LOG_TRC("Pre-reading directory: " << basePath << path << "\n");
     workingdir = opendir((basePath + path).c_str());
 
     while ((currentFile = readdir(workingdir)) != NULL)
@@ -299,7 +336,7 @@ void FileServerRequestHandler::readDirToHash(const std::string &basePath, const
         if (currentFile->d_name[0] == '.')
             continue;
 
-        std::string relPath = path + "/" + currentFile->d_name;
+        const std::string relPath = path + '/' + currentFile->d_name;
         stat ((basePath + relPath).c_str(), &fileStat);
 
         if (S_ISDIR(fileStat.st_mode))
@@ -307,7 +344,7 @@ void FileServerRequestHandler::readDirToHash(const std::string &basePath, const
 
         else if (S_ISREG(fileStat.st_mode))
         {
-            LOG_TRC("Reading file: '" << (basePath + relPath) << " as '" << relPath << "'\n");
+            LOG_TRC("Reading file: '" << basePath << relPath << " as '" << relPath << "'\n");
 
             std::ifstream file(basePath + relPath, std::ios::binary);
 
@@ -318,18 +355,18 @@ void FileServerRequestHandler::readDirToHash(const std::string &basePath, const
             deflateInit2(&strm, Z_DEFAULT_COMPRESSION, Z_DEFLATED, 31, 8, Z_DEFAULT_STRATEGY);
 
             auto buf = std::unique_ptr<char[]>(new char[fileStat.st_size]);
-            std::string compressedFile = "";
-            std::string uncompressedFile = "";
+            std::string compressedFile;
+            compressedFile.reserve(fileStat.st_size);
+            std::string uncompressedFile;
+            uncompressedFile.reserve(fileStat.st_size);
             do {
                 file.read(&buf[0], fileStat.st_size);
                 const long unsigned int size = file.gcount();
                 if (size == 0)
                     break;
 
-                long unsigned int haveComp;
-                long unsigned int compSize = compressBound(size);
-                char *cbuf;
-                cbuf = (char *)calloc(compSize, sizeof(char));
+                const long unsigned int compSize = compressBound(size);
+                char *cbuf = (char *)calloc(compSize, sizeof(char));
 
                 strm.next_in = (unsigned char *)&buf[0];
                 strm.avail_in = size;
@@ -338,16 +375,17 @@ void FileServerRequestHandler::readDirToHash(const std::string &basePath, const
 
                 deflate(&strm, Z_FINISH);
 
-                haveComp = compSize - strm.avail_out;
+                const long unsigned int haveComp = compSize - strm.avail_out;
                 std::string partialcompFile(cbuf, haveComp);
                 std::string partialuncompFile(buf.get(), size);
-                compressedFile = compressedFile + partialcompFile;
-                uncompressedFile = uncompressedFile + partialuncompFile;
+                compressedFile += partialcompFile;
+                uncompressedFile += partialuncompFile;
+                free(cbuf);
 
             } while(true);
 
-            std::pair<std::string, std::string> FilePair(uncompressedFile, compressedFile);
-            FileHash.emplace(relPath, FilePair);
+            FileHash.emplace(relPath, std::make_pair(uncompressedFile, compressedFile));
+            deflateEnd(&strm);
         }
     }
     closedir(workingdir);
@@ -398,35 +436,21 @@ void FileServerRequestHandler::preprocessFile(const HTTPRequest& request, Poco::
     // Is this a file we read at startup - if not; its not for serving.
     const std::string relPath = getRequestPathname(request);
     LOG_DBG("Preprocessing file: " << relPath);
-    if (FileHash.find(relPath) == FileHash.end())
-    {
-        LOG_ERR("File [" << relPath << "] does not exist.");
-
-        // 404 not found
-        std::ostringstream oss;
-        oss << "HTTP/1.1 404\r\n"
-            << "Date: " << Poco::DateTimeFormatter::format(Poco::Timestamp(), Poco::DateTimeFormat::HTTP_FORMAT) << "\r\n"
-            << "User-Agent: " << WOPI_AGENT_STRING << "\r\n"
-            << "Content-Length: 0\r\n"
-            << "\r\n";
-        socket->send(oss.str());
-        return;
-    }
-
     std::string preprocess = *getUncompressedFile(relPath);
 
     HTMLForm form(request, message);
-    const std::string accessToken = form.get("access_token", "");
-    const std::string accessTokenTtl = form.get("access_token_ttl", "");
+    const std::string& accessToken = form.get("access_token", "");
+    const std::string& accessTokenTtl = form.get("access_token_ttl", "");
     LOG_TRC("access_token=" << accessToken << ", access_token_ttl=" << accessTokenTtl);
-    const std::string accessHeader = form.get("access_header", "");
+    const std::string& accessHeader = form.get("access_header", "");
     LOG_TRC("access_header=" << accessHeader);
 
     // Escape bad characters in access token.
     // This is placed directly in javascript in loleaflet.html, we need to make sure
     // that no one can do anything nasty with their clever inputs.
-    std::string escapedAccessToken;
+    std::string escapedAccessToken, escapedAccessHeader;
     Poco::URI::encode(accessToken, "'", escapedAccessToken);
+    Poco::URI::encode(accessHeader, "'", escapedAccessHeader);
 
     unsigned long tokenTtl = 0;
     if (accessToken != "")
@@ -450,6 +474,7 @@ void FileServerRequestHandler::preprocessFile(const HTTPRequest& request, Poco::
 
     Poco::replaceInPlace(preprocess, std::string("%ACCESS_TOKEN%"), escapedAccessToken);
     Poco::replaceInPlace(preprocess, std::string("%ACCESS_TOKEN_TTL%"), std::to_string(tokenTtl));
+    Poco::replaceInPlace(preprocess, std::string("%ACCESS_HEADER%"), escapedAccessHeader);
     Poco::replaceInPlace(preprocess, std::string("%HOST%"), host);
     Poco::replaceInPlace(preprocess, std::string("%VERSION%"), std::string(LOOLWSD_VERSION_HASH));
 
@@ -507,12 +532,9 @@ void FileServerRequestHandler::preprocessFile(const HTTPRequest& request, Poco::
 
     // Keep only the origin, reject everything else
     Poco::URI uriFrameAncestor(frameAncestor);
-    std::string frameAncestorScheme = uriFrameAncestor.getScheme();
-    std::string frameAncestorHost = uriFrameAncestor.getHost();
-
-    if (!frameAncestor.empty() && Util::isValidURIScheme(frameAncestorScheme) && Util::isValidURIHost(frameAncestorHost))
+    if (!frameAncestor.empty() && !uriFrameAncestor.getScheme().empty() && !uriFrameAncestor.getHost().empty())
     {
-        frameAncestor = frameAncestorScheme + "://" + frameAncestorHost + ":" + std::to_string(uriFrameAncestor.getPort());
+        frameAncestor = uriFrameAncestor.getScheme() + "://" + uriFrameAncestor.getHost() + ":" + std::to_string(uriFrameAncestor.getPort());
 
         LOG_TRC("Final frame ancestor: " << frameAncestor);
 
